from protos import ydb_operation_pb2 as _ydb_operation_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class StorageChannelInfo(_message.Message):
    __slots__ = ("storage_channel", "status_flag")
    class StatusFlag(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        STATUS_FLAG_UNSPECIFIED: _ClassVar[StorageChannelInfo.StatusFlag]
        STATUS_FLAG_GREEN: _ClassVar[StorageChannelInfo.StatusFlag]
        STATUS_FLAG_YELLOW_STOP: _ClassVar[StorageChannelInfo.StatusFlag]
        STATUS_FLAG_ORANGE_OUT_SPACE: _ClassVar[StorageChannelInfo.StatusFlag]
    STATUS_FLAG_UNSPECIFIED: StorageChannelInfo.StatusFlag
    STATUS_FLAG_GREEN: StorageChannelInfo.StatusFlag
    STATUS_FLAG_YELLOW_STOP: StorageChannelInfo.StatusFlag
    STATUS_FLAG_ORANGE_OUT_SPACE: StorageChannelInfo.StatusFlag
    STORAGE_CHANNEL_FIELD_NUMBER: _ClassVar[int]
    STATUS_FLAG_FIELD_NUMBER: _ClassVar[int]
    storage_channel: int
    status_flag: StorageChannelInfo.StatusFlag
    def __init__(self, storage_channel: _Optional[int] = ..., status_flag: _Optional[_Union[StorageChannelInfo.StatusFlag, str]] = ...) -> None: ...

class Priorities(_message.Message):
    __slots__ = ()
    class Priority(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        PRIORITY_UNSPECIFIED: _ClassVar[Priorities.Priority]
        PRIORITY_REALTIME: _ClassVar[Priorities.Priority]
        PRIORITY_BACKGROUND: _ClassVar[Priorities.Priority]
    PRIORITY_UNSPECIFIED: Priorities.Priority
    PRIORITY_REALTIME: Priorities.Priority
    PRIORITY_BACKGROUND: Priorities.Priority
    def __init__(self) -> None: ...

class StorageConfig(_message.Message):
    __slots__ = ("channel",)
    class ChannelConfig(_message.Message):
        __slots__ = ("media",)
        MEDIA_FIELD_NUMBER: _ClassVar[int]
        media: str
        def __init__(self, media: _Optional[str] = ...) -> None: ...
    CHANNEL_FIELD_NUMBER: _ClassVar[int]
    channel: _containers.RepeatedCompositeFieldContainer[StorageConfig.ChannelConfig]
    def __init__(self, channel: _Optional[_Iterable[_Union[StorageConfig.ChannelConfig, _Mapping]]] = ...) -> None: ...

class KeyRange(_message.Message):
    __slots__ = ("from_key_inclusive", "from_key_exclusive", "to_key_inclusive", "to_key_exclusive")
    FROM_KEY_INCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    FROM_KEY_EXCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    TO_KEY_INCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    TO_KEY_EXCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    from_key_inclusive: str
    from_key_exclusive: str
    to_key_inclusive: str
    to_key_exclusive: str
    def __init__(self, from_key_inclusive: _Optional[str] = ..., from_key_exclusive: _Optional[str] = ..., to_key_inclusive: _Optional[str] = ..., to_key_exclusive: _Optional[str] = ...) -> None: ...

class AcquireLockRequest(_message.Message):
    __slots__ = ("operation_params", "path", "partition_id")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    PARTITION_ID_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    partition_id: int
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ..., partition_id: _Optional[int] = ...) -> None: ...

class AcquireLockResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class AcquireLockResult(_message.Message):
    __slots__ = ("lock_generation", "node_id")
    LOCK_GENERATION_FIELD_NUMBER: _ClassVar[int]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    lock_generation: int
    node_id: int
    def __init__(self, lock_generation: _Optional[int] = ..., node_id: _Optional[int] = ...) -> None: ...

class ExecuteTransactionRequest(_message.Message):
    __slots__ = ("operation_params", "path", "partition_id", "lock_generation", "commands")
    class Command(_message.Message):
        __slots__ = ("delete_range", "rename", "copy_range", "concat", "write")
        class Rename(_message.Message):
            __slots__ = ("old_key", "new_key")
            OLD_KEY_FIELD_NUMBER: _ClassVar[int]
            NEW_KEY_FIELD_NUMBER: _ClassVar[int]
            old_key: str
            new_key: str
            def __init__(self, old_key: _Optional[str] = ..., new_key: _Optional[str] = ...) -> None: ...
        class Concat(_message.Message):
            __slots__ = ("input_keys", "output_key", "keep_inputs")
            INPUT_KEYS_FIELD_NUMBER: _ClassVar[int]
            OUTPUT_KEY_FIELD_NUMBER: _ClassVar[int]
            KEEP_INPUTS_FIELD_NUMBER: _ClassVar[int]
            input_keys: _containers.RepeatedScalarFieldContainer[str]
            output_key: str
            keep_inputs: bool
            def __init__(self, input_keys: _Optional[_Iterable[str]] = ..., output_key: _Optional[str] = ..., keep_inputs: bool = ...) -> None: ...
        class CopyRange(_message.Message):
            __slots__ = ("range", "prefix_to_remove", "prefix_to_add")
            RANGE_FIELD_NUMBER: _ClassVar[int]
            PREFIX_TO_REMOVE_FIELD_NUMBER: _ClassVar[int]
            PREFIX_TO_ADD_FIELD_NUMBER: _ClassVar[int]
            range: KeyRange
            prefix_to_remove: str
            prefix_to_add: str
            def __init__(self, range: _Optional[_Union[KeyRange, _Mapping]] = ..., prefix_to_remove: _Optional[str] = ..., prefix_to_add: _Optional[str] = ...) -> None: ...
        class Write(_message.Message):
            __slots__ = ("key", "value", "storage_channel", "priority", "tactic")
            class Tactic(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = ()
                TACTIC_UNSPECIFIED: _ClassVar[ExecuteTransactionRequest.Command.Write.Tactic]
                TACTIC_MAX_THROUGHPUT: _ClassVar[ExecuteTransactionRequest.Command.Write.Tactic]
                TACTIC_MIN_LATENCY: _ClassVar[ExecuteTransactionRequest.Command.Write.Tactic]
            TACTIC_UNSPECIFIED: ExecuteTransactionRequest.Command.Write.Tactic
            TACTIC_MAX_THROUGHPUT: ExecuteTransactionRequest.Command.Write.Tactic
            TACTIC_MIN_LATENCY: ExecuteTransactionRequest.Command.Write.Tactic
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            STORAGE_CHANNEL_FIELD_NUMBER: _ClassVar[int]
            PRIORITY_FIELD_NUMBER: _ClassVar[int]
            TACTIC_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: bytes
            storage_channel: int
            priority: Priorities.Priority
            tactic: ExecuteTransactionRequest.Command.Write.Tactic
            def __init__(self, key: _Optional[str] = ..., value: _Optional[bytes] = ..., storage_channel: _Optional[int] = ..., priority: _Optional[_Union[Priorities.Priority, str]] = ..., tactic: _Optional[_Union[ExecuteTransactionRequest.Command.Write.Tactic, str]] = ...) -> None: ...
        class DeleteRange(_message.Message):
            __slots__ = ("range",)
            RANGE_FIELD_NUMBER: _ClassVar[int]
            range: KeyRange
            def __init__(self, range: _Optional[_Union[KeyRange, _Mapping]] = ...) -> None: ...
        DELETE_RANGE_FIELD_NUMBER: _ClassVar[int]
        RENAME_FIELD_NUMBER: _ClassVar[int]
        COPY_RANGE_FIELD_NUMBER: _ClassVar[int]
        CONCAT_FIELD_NUMBER: _ClassVar[int]
        WRITE_FIELD_NUMBER: _ClassVar[int]
        delete_range: ExecuteTransactionRequest.Command.DeleteRange
        rename: ExecuteTransactionRequest.Command.Rename
        copy_range: ExecuteTransactionRequest.Command.CopyRange
        concat: ExecuteTransactionRequest.Command.Concat
        write: ExecuteTransactionRequest.Command.Write
        def __init__(self, delete_range: _Optional[_Union[ExecuteTransactionRequest.Command.DeleteRange, _Mapping]] = ..., rename: _Optional[_Union[ExecuteTransactionRequest.Command.Rename, _Mapping]] = ..., copy_range: _Optional[_Union[ExecuteTransactionRequest.Command.CopyRange, _Mapping]] = ..., concat: _Optional[_Union[ExecuteTransactionRequest.Command.Concat, _Mapping]] = ..., write: _Optional[_Union[ExecuteTransactionRequest.Command.Write, _Mapping]] = ...) -> None: ...
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    PARTITION_ID_FIELD_NUMBER: _ClassVar[int]
    LOCK_GENERATION_FIELD_NUMBER: _ClassVar[int]
    COMMANDS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    partition_id: int
    lock_generation: int
    commands: _containers.RepeatedCompositeFieldContainer[ExecuteTransactionRequest.Command]
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ..., partition_id: _Optional[int] = ..., lock_generation: _Optional[int] = ..., commands: _Optional[_Iterable[_Union[ExecuteTransactionRequest.Command, _Mapping]]] = ...) -> None: ...

class ExecuteTransactionResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ExecuteTransactionResult(_message.Message):
    __slots__ = ("storage_channel_info", "node_id")
    STORAGE_CHANNEL_INFO_FIELD_NUMBER: _ClassVar[int]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    storage_channel_info: _containers.RepeatedCompositeFieldContainer[StorageChannelInfo]
    node_id: int
    def __init__(self, storage_channel_info: _Optional[_Iterable[_Union[StorageChannelInfo, _Mapping]]] = ..., node_id: _Optional[int] = ...) -> None: ...

class ReadRequest(_message.Message):
    __slots__ = ("operation_params", "path", "partition_id", "lock_generation", "key", "offset", "size", "limit_bytes", "priority")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    PARTITION_ID_FIELD_NUMBER: _ClassVar[int]
    LOCK_GENERATION_FIELD_NUMBER: _ClassVar[int]
    KEY_FIELD_NUMBER: _ClassVar[int]
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    SIZE_FIELD_NUMBER: _ClassVar[int]
    LIMIT_BYTES_FIELD_NUMBER: _ClassVar[int]
    PRIORITY_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    partition_id: int
    lock_generation: int
    key: str
    offset: int
    size: int
    limit_bytes: int
    priority: Priorities.Priority
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ..., partition_id: _Optional[int] = ..., lock_generation: _Optional[int] = ..., key: _Optional[str] = ..., offset: _Optional[int] = ..., size: _Optional[int] = ..., limit_bytes: _Optional[int] = ..., priority: _Optional[_Union[Priorities.Priority, str]] = ...) -> None: ...

class ReadResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ReadResult(_message.Message):
    __slots__ = ("requested_key", "requested_offset", "requested_size", "value", "is_overrun", "node_id")
    REQUESTED_KEY_FIELD_NUMBER: _ClassVar[int]
    REQUESTED_OFFSET_FIELD_NUMBER: _ClassVar[int]
    REQUESTED_SIZE_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    IS_OVERRUN_FIELD_NUMBER: _ClassVar[int]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    requested_key: str
    requested_offset: int
    requested_size: int
    value: bytes
    is_overrun: bool
    node_id: int
    def __init__(self, requested_key: _Optional[str] = ..., requested_offset: _Optional[int] = ..., requested_size: _Optional[int] = ..., value: _Optional[bytes] = ..., is_overrun: bool = ..., node_id: _Optional[int] = ...) -> None: ...

class ReadRangeRequest(_message.Message):
    __slots__ = ("operation_params", "path", "partition_id", "lock_generation", "range", "limit_bytes", "priority")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    PARTITION_ID_FIELD_NUMBER: _ClassVar[int]
    LOCK_GENERATION_FIELD_NUMBER: _ClassVar[int]
    RANGE_FIELD_NUMBER: _ClassVar[int]
    LIMIT_BYTES_FIELD_NUMBER: _ClassVar[int]
    PRIORITY_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    partition_id: int
    lock_generation: int
    range: KeyRange
    limit_bytes: int
    priority: Priorities.Priority
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ..., partition_id: _Optional[int] = ..., lock_generation: _Optional[int] = ..., range: _Optional[_Union[KeyRange, _Mapping]] = ..., limit_bytes: _Optional[int] = ..., priority: _Optional[_Union[Priorities.Priority, str]] = ...) -> None: ...

class ReadRangeResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ReadRangeResult(_message.Message):
    __slots__ = ("pair", "is_overrun", "node_id")
    class KeyValuePair(_message.Message):
        __slots__ = ("key", "value", "creation_unix_time", "storage_channel")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        CREATION_UNIX_TIME_FIELD_NUMBER: _ClassVar[int]
        STORAGE_CHANNEL_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bytes
        creation_unix_time: int
        storage_channel: int
        def __init__(self, key: _Optional[str] = ..., value: _Optional[bytes] = ..., creation_unix_time: _Optional[int] = ..., storage_channel: _Optional[int] = ...) -> None: ...
    PAIR_FIELD_NUMBER: _ClassVar[int]
    IS_OVERRUN_FIELD_NUMBER: _ClassVar[int]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    pair: _containers.RepeatedCompositeFieldContainer[ReadRangeResult.KeyValuePair]
    is_overrun: bool
    node_id: int
    def __init__(self, pair: _Optional[_Iterable[_Union[ReadRangeResult.KeyValuePair, _Mapping]]] = ..., is_overrun: bool = ..., node_id: _Optional[int] = ...) -> None: ...

class ListRangeRequest(_message.Message):
    __slots__ = ("operation_params", "path", "partition_id", "lock_generation", "range", "limit_bytes")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    PARTITION_ID_FIELD_NUMBER: _ClassVar[int]
    LOCK_GENERATION_FIELD_NUMBER: _ClassVar[int]
    RANGE_FIELD_NUMBER: _ClassVar[int]
    LIMIT_BYTES_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    partition_id: int
    lock_generation: int
    range: KeyRange
    limit_bytes: int
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ..., partition_id: _Optional[int] = ..., lock_generation: _Optional[int] = ..., range: _Optional[_Union[KeyRange, _Mapping]] = ..., limit_bytes: _Optional[int] = ...) -> None: ...

class ListRangeResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ListRangeResult(_message.Message):
    __slots__ = ("key", "is_overrun", "node_id")
    class KeyInfo(_message.Message):
        __slots__ = ("key", "value_size", "creation_unix_time", "storage_channel")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_SIZE_FIELD_NUMBER: _ClassVar[int]
        CREATION_UNIX_TIME_FIELD_NUMBER: _ClassVar[int]
        STORAGE_CHANNEL_FIELD_NUMBER: _ClassVar[int]
        key: str
        value_size: int
        creation_unix_time: int
        storage_channel: int
        def __init__(self, key: _Optional[str] = ..., value_size: _Optional[int] = ..., creation_unix_time: _Optional[int] = ..., storage_channel: _Optional[int] = ...) -> None: ...
    KEY_FIELD_NUMBER: _ClassVar[int]
    IS_OVERRUN_FIELD_NUMBER: _ClassVar[int]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    key: _containers.RepeatedCompositeFieldContainer[ListRangeResult.KeyInfo]
    is_overrun: bool
    node_id: int
    def __init__(self, key: _Optional[_Iterable[_Union[ListRangeResult.KeyInfo, _Mapping]]] = ..., is_overrun: bool = ..., node_id: _Optional[int] = ...) -> None: ...

class GetStorageChannelStatusRequest(_message.Message):
    __slots__ = ("operation_params", "path", "partition_id", "lock_generation", "storage_channel")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    PARTITION_ID_FIELD_NUMBER: _ClassVar[int]
    LOCK_GENERATION_FIELD_NUMBER: _ClassVar[int]
    STORAGE_CHANNEL_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    partition_id: int
    lock_generation: int
    storage_channel: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ..., partition_id: _Optional[int] = ..., lock_generation: _Optional[int] = ..., storage_channel: _Optional[_Iterable[int]] = ...) -> None: ...

class GetStorageChannelStatusResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class GetStorageChannelStatusResult(_message.Message):
    __slots__ = ("storage_channel_info", "node_id")
    STORAGE_CHANNEL_INFO_FIELD_NUMBER: _ClassVar[int]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    storage_channel_info: _containers.RepeatedCompositeFieldContainer[StorageChannelInfo]
    node_id: int
    def __init__(self, storage_channel_info: _Optional[_Iterable[_Union[StorageChannelInfo, _Mapping]]] = ..., node_id: _Optional[int] = ...) -> None: ...

class CreateVolumeRequest(_message.Message):
    __slots__ = ("operation_params", "path", "partition_count", "storage_config")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    PARTITION_COUNT_FIELD_NUMBER: _ClassVar[int]
    STORAGE_CONFIG_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    partition_count: int
    storage_config: StorageConfig
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ..., partition_count: _Optional[int] = ..., storage_config: _Optional[_Union[StorageConfig, _Mapping]] = ...) -> None: ...

class CreateVolumeResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class CreateVolumeResult(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class DropVolumeRequest(_message.Message):
    __slots__ = ("operation_params", "path")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ...) -> None: ...

class DropVolumeResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class DropVolumeResult(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class AlterVolumeRequest(_message.Message):
    __slots__ = ("operation_params", "path", "alter_partition_count", "storage_config")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    ALTER_PARTITION_COUNT_FIELD_NUMBER: _ClassVar[int]
    STORAGE_CONFIG_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    alter_partition_count: int
    storage_config: StorageConfig
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ..., alter_partition_count: _Optional[int] = ..., storage_config: _Optional[_Union[StorageConfig, _Mapping]] = ...) -> None: ...

class AlterVolumeResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class AlterVolumeResult(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class DescribeVolumeRequest(_message.Message):
    __slots__ = ("operation_params", "path")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ...) -> None: ...

class DescribeVolumeResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class DescribeVolumeResult(_message.Message):
    __slots__ = ("path", "partition_count")
    PATH_FIELD_NUMBER: _ClassVar[int]
    PARTITION_COUNT_FIELD_NUMBER: _ClassVar[int]
    path: str
    partition_count: int
    def __init__(self, path: _Optional[str] = ..., partition_count: _Optional[int] = ...) -> None: ...

class ListLocalPartitionsRequest(_message.Message):
    __slots__ = ("operation_params", "path", "node_id")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    node_id: int
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ..., node_id: _Optional[int] = ...) -> None: ...

class ListLocalPartitionsResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ListLocalPartitionsResult(_message.Message):
    __slots__ = ("path", "node_id", "partition_ids")
    PATH_FIELD_NUMBER: _ClassVar[int]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    PARTITION_IDS_FIELD_NUMBER: _ClassVar[int]
    path: str
    node_id: int
    partition_ids: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, path: _Optional[str] = ..., node_id: _Optional[int] = ..., partition_ids: _Optional[_Iterable[int]] = ...) -> None: ...



import asyncio
import functools
import inspect
import random
import time

from . import issues
from ._errors import check_retriable_error


class BackoffSettings(object):
    def __init__(self, ceiling=6, slot_duration=0.001, uncertain_ratio=0.5):
        self.ceiling = ceiling
        self.slot_duration = slot_duration
        self.uncertain_ratio = uncertain_ratio

    def calc_timeout(self, retry_number):
        slots_count = 1 << min(retry_number, self.ceiling)
        max_duration_ms = slots_count * self.slot_duration * 1000.0
        # duration_ms = random.random() * max_duration_ms * uncertain_ratio) + max_duration_ms * (1 - uncertain_ratio)
        duration_ms = max_duration_ms * (random.random() * self.uncertain_ratio + 1.0 - self.uncertain_ratio)
        return duration_ms / 1000.0


class RetrySettings(object):
    def __init__(
        self,
        max_retries=10,
        max_session_acquire_timeout=None,
        on_ydb_error_callback=None,
        backoff_ceiling=6,
        backoff_slot_duration=1,
        get_session_client_timeout=5,
        fast_backoff_settings=None,
        slow_backoff_settings=None,
        idempotent=False,
        retry_cancelled=False,
    ):
        self.max_retries = max_retries
        self.max_session_acquire_timeout = max_session_acquire_timeout
        self.on_ydb_error_callback = (lambda e: None) if on_ydb_error_callback is None else on_ydb_error_callback
        self.fast_backoff = BackoffSettings(10, 0.005) if fast_backoff_settings is None else fast_backoff_settings
        self.slow_backoff = (
            BackoffSettings(backoff_ceiling, backoff_slot_duration)
            if slow_backoff_settings is None
            else slow_backoff_settings
        )
        self.retry_not_found = True
        self.idempotent = idempotent
        self.retry_internal_error = True
        self.retry_cancelled = retry_cancelled
        self.unknown_error_handler = lambda e: None
        self.get_session_client_timeout = get_session_client_timeout
        if max_session_acquire_timeout is not None:
            self.get_session_client_timeout = min(self.max_session_acquire_timeout, self.get_session_client_timeout)

    def with_fast_backoff(self, backoff_settings):
        self.fast_backoff = backoff_settings
        return self

    def with_slow_backoff(self, backoff_settings):
        self.slow_backoff = backoff_settings
        return self


class YdbRetryOperationSleepOpt(object):
    def __init__(self, timeout):
        self.timeout = timeout

    def __eq__(self, other):
        return type(self) == type(other) and self.timeout == other.timeout

    def __repr__(self):
        return "YdbRetryOperationSleepOpt(%s)" % self.timeout


class YdbRetryOperationFinalResult(object):
    def __init__(self, result):
        self.result = result
        self.exc = None

    def __eq__(self, other):
        return type(self) == type(other) and self.result == other.result and self.exc == other.exc

    def __repr__(self):
        return "YdbRetryOperationFinalResult(%s, exc=%s)" % (self.result, self.exc)

    def set_exception(self, exc):
        self.exc = exc


def retry_operation_impl(callee, retry_settings=None, *args, **kwargs):
    retry_settings = RetrySettings() if retry_settings is None else retry_settings
    status = None

    for attempt in range(retry_settings.max_retries + 1):
        try:
            result = YdbRetryOperationFinalResult(callee(*args, **kwargs))
            yield result

            if result.exc is not None:
                raise result.exc

        except issues.Error as e:
            status = e
            retry_settings.on_ydb_error_callback(e)

            retriable_info = check_retriable_error(e, retry_settings, attempt)
            if not retriable_info.is_retriable:
                raise

            skip_yield_error_types = [
                issues.Aborted,
                issues.BadSession,
                issues.NotFound,
                issues.InternalError,
            ]

            yield_sleep = True
            for t in skip_yield_error_types:
                if isinstance(e, t):
                    yield_sleep = False

            if yield_sleep:
                yield YdbRetryOperationSleepOpt(retriable_info.sleep_timeout_seconds)

        except Exception as e:
            # you should provide your own handler you want
            retry_settings.unknown_error_handler(e)
            raise

    raise status


def retry_operation_sync(callee, retry_settings=None, *args, **kwargs):
    opt_generator = retry_operation_impl(callee, retry_settings, *args, **kwargs)
    for next_opt in opt_generator:
        if isinstance(next_opt, YdbRetryOperationSleepOpt):
            time.sleep(next_opt.timeout)
        else:
            return next_opt.result


async def retry_operation_async(callee, retry_settings=None, *args, **kwargs):  # pylint: disable=W1113
    """
    The retry operation helper can be used to retry a coroutine that raises YDB specific
    exceptions.

    :param callee: A coroutine to retry.
    :param retry_settings: An instance of ydb.RetrySettings that describes how the coroutine
    should be retried. If None, default instance of retry settings will be used.
    :param args: A tuple with positional arguments to be passed into the coroutine.
    :param kwargs: A dictionary with keyword arguments to be passed into the coroutine.

    Returns awaitable result of coroutine. If retries are not succussful exception is raised.
    """
    opt_generator = retry_operation_impl(callee, retry_settings, *args, **kwargs)
    for next_opt in opt_generator:
        if isinstance(next_opt, YdbRetryOperationSleepOpt):
            await asyncio.sleep(next_opt.timeout)
        else:
            try:
                return await next_opt.result
            except BaseException as e:  # pylint: disable=W0703
                next_opt.set_exception(e)


def ydb_retry(
    max_retries=10,
    max_session_acquire_timeout=None,
    on_ydb_error_callback=None,
    backoff_ceiling=6,
    backoff_slot_duration=1,
    get_session_client_timeout=5,
    fast_backoff_settings=None,
    slow_backoff_settings=None,
    idempotent=False,
    retry_cancelled=False,
):
    """
    Decorator for automatic function retry in case of YDB errors.

    Supports both synchronous and asynchronous functions.

    :param max_retries: Maximum number of retries (default: 10)
    :param max_session_acquire_timeout: Maximum session acquisition timeout (default: None)
    :param on_ydb_error_callback: Callback for handling YDB errors (default: None)
    :param backoff_ceiling: Ceiling for backoff algorithm (default: 6)
    :param backoff_slot_duration: Slot duration for backoff (default: 1)
    :param get_session_client_timeout: Session client timeout (default: 5)
    :param fast_backoff_settings: Fast backoff settings (default: None)
    :param slow_backoff_settings: Slow backoff settings (default: None)
    :param idempotent: Whether the operation is idempotent (default: False)
    :param retry_cancelled: Whether to retry cancelled operations (default: False)
    """

    def decorator(func):
        retry_settings = RetrySettings(
            max_retries=max_retries,
            max_session_acquire_timeout=max_session_acquire_timeout,
            on_ydb_error_callback=on_ydb_error_callback,
            backoff_ceiling=backoff_ceiling,
            backoff_slot_duration=backoff_slot_duration,
            get_session_client_timeout=get_session_client_timeout,
            fast_backoff_settings=fast_backoff_settings,
            slow_backoff_settings=slow_backoff_settings,
            idempotent=idempotent,
            retry_cancelled=retry_cancelled,
        )

        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def async_wrapper(*args, **kwargs):
                return await retry_operation_async(func, retry_settings, *args, **kwargs)

            return async_wrapper
        else:

            @functools.wraps(func)
            def sync_wrapper(*args, **kwargs):
                return retry_operation_sync(func, retry_settings, *args, **kwargs)

            return sync_wrapper

    return decorator



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_federation_discovery_v1.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v5.protos import ydb_federation_discovery_pb2 as protos_dot_ydb__federation__discovery__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!ydb_federation_discovery_v1.proto\x12\x1aYdb.FederationDiscovery.V1\x1a%protos/ydb_federation_discovery.proto2\xab\x01\n\x1a\x46\x65\x64\x65rationDiscoveryService\x12\x8c\x01\n\x17ListFederationDatabases\x12\x37.Ydb.FederationDiscovery.ListFederationDatabasesRequest\x1a\x38.Ydb.FederationDiscovery.ListFederationDatabasesResponseBl\n&tech.ydb.proto.federation.discovery.v1ZBgithub.com/ydb-platform/ydb-go-genproto/Ydb_FederationDiscovery_V1b\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb_federation_discovery_v1_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n&tech.ydb.proto.federation.discovery.v1ZBgithub.com/ydb-platform/ydb-go-genproto/Ydb_FederationDiscovery_V1'
  _FEDERATIONDISCOVERYSERVICE._serialized_start=105
  _FEDERATIONDISCOVERYSERVICE._serialized_end=276
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

__all__ = [
    "TopicClient",
    "TopicClientAsyncIO",
    "TopicClientSettings",
    "TopicCodec",
    "TopicConsumer",
    "TopicAlterConsumer",
    "TopicAlterAutoPartitioningSettings",
    "TopicAutoPartitioningSettings",
    "TopicAutoPartitioningStrategy",
    "TopicDescription",
    "TopicError",
    "TopicMeteringMode",
    "TopicReader",
    "TopicReaderAsyncIO",
    "TopicReaderBatch",
    "TopicReaderEvents",
    "TopicReaderMessage",
    "TopicReaderSelector",
    "TopicReaderSettings",
    "TopicReaderUnexpectedCodecError",
    "TopicReaderPartitionExpiredError",
    "TopicStatWindow",
    "TopicWriteResult",
    "TopicWriter",
    "TopicWriterAsyncIO",
    "TopicTxWriter",
    "TopicTxWriterAsyncIO",
    "TopicWriterInitInfo",
    "TopicWriterMessage",
    "TopicWriterSettings",
]

import concurrent.futures
import datetime
from dataclasses import dataclass
import logging
from typing import List, Union, Mapping, Optional, Dict, Callable

from . import aio, Credentials, _apis, issues

from . import driver

from ._topic_reader import events as TopicReaderEvents

from ._topic_reader.datatypes import (
    PublicBatch as TopicReaderBatch,
    PublicMessage as TopicReaderMessage,
)

from ._topic_reader.topic_reader import (
    PublicReaderSettings as TopicReaderSettings,
    PublicTopicSelector as TopicReaderSelector,
)

from ._topic_reader.topic_reader_sync import (
    TopicReaderSync as TopicReader,
)

from ._topic_reader.topic_reader_asyncio import (
    PublicAsyncIOReader as TopicReaderAsyncIO,
    PublicTopicReaderPartitionExpiredError as TopicReaderPartitionExpiredError,
    PublicTopicReaderUnexpectedCodecError as TopicReaderUnexpectedCodecError,
)

from ._topic_writer.topic_writer import (  # noqa: F401
    PublicWriterSettings as TopicWriterSettings,
    PublicMessage as TopicWriterMessage,
    RetryPolicy as TopicWriterRetryPolicy,
    PublicWriterInitInfo as TopicWriterInitInfo,
    PublicWriteResult as TopicWriteResult,
)

from ydb._topic_writer.topic_writer_asyncio import TxWriterAsyncIO as TopicTxWriterAsyncIO
from ydb._topic_writer.topic_writer_asyncio import WriterAsyncIO as TopicWriterAsyncIO
from ._topic_writer.topic_writer_sync import WriterSync as TopicWriter
from ._topic_writer.topic_writer_sync import TxWriterSync as TopicTxWriter

from ._topic_common.common import (
    wrap_operation as _wrap_operation,
    create_result_wrapper as _create_result_wrapper,
)

from ._grpc.grpcwrapper import ydb_topic as _ydb_topic
from ._grpc.grpcwrapper import ydb_topic_public_types as _ydb_topic_public_types
from ._grpc.grpcwrapper.ydb_topic_public_types import (  # noqa: F401
    PublicDescribeTopicResult as TopicDescription,
    PublicMultipleWindowsStat as TopicStatWindow,
    PublicPartitionStats as TopicPartitionStats,
    PublicCodec as TopicCodec,
    PublicConsumer as TopicConsumer,
    PublicAlterConsumer as TopicAlterConsumer,
    PublicMeteringMode as TopicMeteringMode,
    PublicAutoPartitioningStrategy as TopicAutoPartitioningStrategy,
    PublicAutoPartitioningSettings as TopicAutoPartitioningSettings,
    PublicAlterAutoPartitioningSettings as TopicAlterAutoPartitioningSettings,
)

from .retries import ydb_retry

logger = logging.getLogger(__name__)


class TopicClientAsyncIO:
    _closed: bool
    _driver: aio.Driver
    _credentials: Union[Credentials, None]
    _settings: TopicClientSettings
    _executor: concurrent.futures.Executor

    def __init__(self, driver: aio.Driver, settings: Optional[TopicClientSettings] = None):
        if not settings:
            settings = TopicClientSettings()
        self._closed = False
        self._driver = driver
        self._settings = settings
        self._executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=settings.encode_decode_threads_count,
            thread_name_prefix="topic_asyncio_executor",
        )

    def __del__(self):
        if not self._closed:
            try:
                logger.debug("Topic client was not closed properly. Consider using method close().")
                self.close()
            except BaseException:
                logger.warning("Something went wrong during topic client close in __del__")

    async def create_topic(
        self,
        path: str,
        min_active_partitions: Optional[int] = None,
        max_active_partitions: Optional[int] = None,
        partition_count_limit: Optional[int] = None,
        retention_period: Optional[datetime.timedelta] = None,
        retention_storage_mb: Optional[int] = None,
        supported_codecs: Optional[List[Union[TopicCodec, int]]] = None,
        partition_write_speed_bytes_per_second: Optional[int] = None,
        partition_write_burst_bytes: Optional[int] = None,
        attributes: Optional[Dict[str, str]] = None,
        consumers: Optional[List[Union[TopicConsumer, str]]] = None,
        metering_mode: Optional[TopicMeteringMode] = None,
        auto_partitioning_settings: Optional[TopicAutoPartitioningSettings] = None,
    ):
        """
        create topic command

        :param path: full path to topic
        :param min_active_partitions: Minimum partition count auto merge would stop working at.
        :param partition_count_limit: Limit for total partition count, including active (open for write)
            and read-only partitions.
        :param retention_period: How long data in partition should be stored
        :param retention_storage_mb: How much data in partition should be stored
        :param supported_codecs: List of allowed codecs for writers. Writes with codec not from this list are forbidden.
            Empty list mean disable codec compatibility checks for the topic.
        :param partition_write_speed_bytes_per_second: Partition write speed in bytes per second
        :param partition_write_burst_bytes: Burst size for write in partition, in bytes
        :param attributes: User and server attributes of topic.
            Server attributes starts from "_" and will be validated by server.
        :param consumers: List of consumers for this topic
        :param metering_mode: Metering mode for the topic in a serverless database
        """
        logger.debug("Create topic request: path=%s", path)
        args = locals().copy()
        del args["self"]
        req = _ydb_topic_public_types.CreateTopicRequestParams(**args)
        req = _ydb_topic.CreateTopicRequest.from_public(req)
        await self._driver(
            req.to_proto(),
            _apis.TopicService.Stub,
            _apis.TopicService.CreateTopic,
            _wrap_operation,
        )

    async def alter_topic(
        self,
        path: str,
        set_min_active_partitions: Optional[int] = None,
        set_max_active_partitions: Optional[int] = None,
        set_partition_count_limit: Optional[int] = None,
        add_consumers: Optional[List[Union[TopicConsumer, str]]] = None,
        alter_consumers: Optional[List[Union[TopicAlterConsumer, str]]] = None,
        drop_consumers: Optional[List[str]] = None,
        alter_attributes: Optional[Dict[str, str]] = None,
        set_metering_mode: Optional[TopicMeteringMode] = None,
        set_partition_write_speed_bytes_per_second: Optional[int] = None,
        set_partition_write_burst_bytes: Optional[int] = None,
        set_retention_period: Optional[datetime.timedelta] = None,
        set_retention_storage_mb: Optional[int] = None,
        set_supported_codecs: Optional[List[Union[TopicCodec, int]]] = None,
        alter_auto_partitioning_settings: Optional[TopicAlterAutoPartitioningSettings] = None,
    ):
        """
        alter topic command

        :param path: full path to topic
        :param set_min_active_partitions: Minimum partition count auto merge would stop working at.
        :param set_partition_count_limit: Limit for total partition count, including active (open for write)
            and read-only partitions.
        :param add_consumers: List of consumers for this topic to add
        :param alter_consumers: List of consumers for this topic to alter
        :param drop_consumers: List of consumer names for this topic to drop
        :param alter_attributes: User and server attributes of topic.
            Server attributes starts from "_" and will be validated by server.
        :param set_metering_mode: Metering mode for the topic in a serverless database
        :param set_partition_write_speed_bytes_per_second: Partition write speed in bytes per second
        :param set_partition_write_burst_bytes: Burst size for write in partition, in bytes
        :param set_retention_period: How long data in partition should be stored
        :param set_retention_storage_mb: How much data in partition should be stored
        :param set_supported_codecs: List of allowed codecs for writers. Writes with codec not from this list are forbidden.
            Empty list mean disable codec compatibility checks for the topic.
        """
        logger.debug("Alter topic request: path=%s", path)
        args = locals().copy()
        del args["self"]
        req = _ydb_topic_public_types.AlterTopicRequestParams(**args)
        req = _ydb_topic.AlterTopicRequest.from_public(req)
        await self._driver(
            req.to_proto(),
            _apis.TopicService.Stub,
            _apis.TopicService.AlterTopic,
            _wrap_operation,
        )

    async def describe_topic(self, path: str, include_stats: bool = False) -> TopicDescription:
        logger.debug("Describe topic request: path=%s", path)
        args = locals().copy()
        del args["self"]
        req = _ydb_topic_public_types.DescribeTopicRequestParams(**args)
        res = await self._driver(
            req.to_proto(),
            _apis.TopicService.Stub,
            _apis.TopicService.DescribeTopic,
            _create_result_wrapper(_ydb_topic.DescribeTopicResult),
        )  # type: _ydb_topic.DescribeTopicResult
        return res.to_public()

    async def drop_topic(self, path: str):
        logger.debug("Drop topic request: path=%s", path)
        req = _ydb_topic_public_types.DropTopicRequestParams(path=path)
        await self._driver(
            req.to_proto(),
            _apis.TopicService.Stub,
            _apis.TopicService.DropTopic,
            _wrap_operation,
        )

    def reader(
        self,
        topic: Union[str, TopicReaderSelector, List[Union[str, TopicReaderSelector]]],
        consumer: Optional[str],
        buffer_size_bytes: int = 50 * 1024 * 1024,
        # decoders: map[codec_code] func(encoded_bytes)->decoded_bytes
        # the func will be called from multiply threads in parallel
        decoders: Union[Mapping[int, Callable[[bytes], bytes]], None] = None,
        # custom decoder executor for call builtin and custom decoders. If None - use shared executor pool.
        # if max_worker in the executor is 1 - then decoders will be called from the thread without parallel
        decoder_executor: Optional[concurrent.futures.Executor] = None,
        auto_partitioning_support: Optional[bool] = True,  # Auto partitioning feature flag. Default - True.
        event_handler: Optional[TopicReaderEvents.EventHandler] = None,
    ) -> TopicReaderAsyncIO:

        logger.debug("Create reader for topic=%s consumer=%s", topic, consumer)

        if not decoder_executor:
            decoder_executor = self._executor

        args = locals().copy()
        del args["self"]

        if consumer == "":
            raise issues.Error(
                "Consumer name could not be empty! To use reader without consumer specify consumer as None."
            )

        if consumer is None:
            if not isinstance(topic, TopicReaderSelector) or topic.partitions is None:
                raise issues.Error(
                    "To use reader without consumer it is required to specify partition_ids in topic selector."
                )

            if event_handler is None:
                raise issues.Error(
                    "To use reader without consumer it is required to specify event_handler with "
                    "on_partition_get_start_offset method."
                )

        settings = TopicReaderSettings(**args)

        return TopicReaderAsyncIO(self._driver, settings, _parent=self)

    def writer(
        self,
        topic,
        *,
        producer_id: Optional[str] = None,  # default - random
        session_metadata: Mapping[str, str] = None,
        partition_id: Union[int, None] = None,
        auto_seqno: bool = True,
        auto_created_at: bool = True,
        codec: Optional[TopicCodec] = None,  # default mean auto-select
        # encoders: map[codec_code] func(encoded_bytes)->decoded_bytes
        # the func will be called from multiply threads in parallel.
        encoders: Optional[Mapping[_ydb_topic_public_types.PublicCodec, Callable[[bytes], bytes]]] = None,
        # custom encoder executor for call builtin and custom decoders. If None - use shared executor pool.
        # If max_worker in the executor is 1 - then encoders will be called from the thread without parallel.
        encoder_executor: Optional[concurrent.futures.Executor] = None,
    ) -> TopicWriterAsyncIO:
        logger.debug("Create writer for topic=%s producer_id=%s", topic, producer_id)
        args = locals().copy()
        del args["self"]

        settings = TopicWriterSettings(**args)

        if not settings.encoder_executor:
            settings.encoder_executor = self._executor

        return TopicWriterAsyncIO(self._driver, settings, _client=self)

    def tx_writer(
        self,
        tx,
        topic,
        *,
        producer_id: Optional[str] = None,  # default - random
        session_metadata: Mapping[str, str] = None,
        partition_id: Union[int, None] = None,
        auto_seqno: bool = True,
        auto_created_at: bool = True,
        codec: Optional[TopicCodec] = None,  # default mean auto-select
        # encoders: map[codec_code] func(encoded_bytes)->decoded_bytes
        # the func will be called from multiply threads in parallel.
        encoders: Optional[Mapping[_ydb_topic_public_types.PublicCodec, Callable[[bytes], bytes]]] = None,
        # custom encoder executor for call builtin and custom decoders. If None - use shared executor pool.
        # If max_worker in the executor is 1 - then encoders will be called from the thread without parallel.
        encoder_executor: Optional[concurrent.futures.Executor] = None,
    ) -> TopicTxWriterAsyncIO:
        logger.debug("Create tx writer for topic=%s tx=%s", topic, tx)
        args = locals().copy()
        del args["self"]
        del args["tx"]

        settings = TopicWriterSettings(**args)

        if not settings.encoder_executor:
            settings.encoder_executor = self._executor

        return TopicTxWriterAsyncIO(tx=tx, driver=self._driver, settings=settings, _client=self)

    @ydb_retry(retry_cancelled=True, idempotent=True)
    async def commit_offset(
        self, path: str, consumer: str, partition_id: int, offset: int, read_session_id: Optional[str] = None
    ) -> None:
        logger.debug(
            "Commit offset: path=%s partition_id=%s offset=%s consumer=%s",
            path,
            partition_id,
            offset,
            consumer,
        )
        req = _ydb_topic.CommitOffsetRequest(
            path=path,
            consumer=consumer,
            partition_id=partition_id,
            offset=offset,
            read_session_id=read_session_id,
        )

        await self._driver(
            req.to_proto(),
            _apis.TopicService.Stub,
            _apis.TopicService.CommitOffset,
            _wrap_operation,
        )

    def close(self):
        if self._closed:
            return

        logger.debug("Close topic client")
        self._closed = True
        self._executor.shutdown(wait=False)

    def _check_closed(self):
        if not self._closed:
            return

        raise issues.Error("Topic client closed")


class TopicClient:
    _closed: bool
    _driver: driver.Driver
    _credentials: Union[Credentials, None]
    _settings: TopicClientSettings
    _executor: concurrent.futures.Executor

    def __init__(self, driver: driver.Driver, settings: Optional[TopicClientSettings]):
        if not settings:
            settings = TopicClientSettings()

        self._closed = False
        self._driver = driver
        self._settings = settings
        self._executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=settings.encode_decode_threads_count,
            thread_name_prefix="topic_asyncio_executor",
        )

    def __del__(self):
        if not self._closed:
            try:
                logger.warning("Topic client was not closed properly. Consider using method close().")
                self.close()
            except BaseException:
                logger.warning("Something went wrong during topic client close in __del__")

    def create_topic(
        self,
        path: str,
        min_active_partitions: Optional[int] = None,
        max_active_partitions: Optional[int] = None,
        partition_count_limit: Optional[int] = None,
        retention_period: Optional[datetime.timedelta] = None,
        retention_storage_mb: Optional[int] = None,
        supported_codecs: Optional[List[Union[TopicCodec, int]]] = None,
        partition_write_speed_bytes_per_second: Optional[int] = None,
        partition_write_burst_bytes: Optional[int] = None,
        attributes: Optional[Dict[str, str]] = None,
        consumers: Optional[List[Union[TopicConsumer, str]]] = None,
        metering_mode: Optional[TopicMeteringMode] = None,
        auto_partitioning_settings: Optional[TopicAutoPartitioningSettings] = None,
    ):
        """
        create topic command

        :param path: full path to topic
        :param min_active_partitions: Minimum partition count auto merge would stop working at.
        :param partition_count_limit: Limit for total partition count, including active (open for write)
            and read-only partitions.
        :param retention_period: How long data in partition should be stored
        :param retention_storage_mb: How much data in partition should be stored
        :param supported_codecs: List of allowed codecs for writers. Writes with codec not from this list are forbidden.
            Empty list mean disable codec compatibility checks for the topic.
        :param partition_write_speed_bytes_per_second: Partition write speed in bytes per second
        :param partition_write_burst_bytes: Burst size for write in partition, in bytes
        :param attributes: User and server attributes of topic.
            Server attributes starts from "_" and will be validated by server.
        :param consumers: List of consumers for this topic
        :param metering_mode: Metering mode for the topic in a serverless database
        """
        logger.debug("Create topic request: path=%s", path)
        args = locals().copy()
        del args["self"]
        self._check_closed()

        req = _ydb_topic_public_types.CreateTopicRequestParams(**args)
        req = _ydb_topic.CreateTopicRequest.from_public(req)
        self._driver(
            req.to_proto(),
            _apis.TopicService.Stub,
            _apis.TopicService.CreateTopic,
            _wrap_operation,
        )

    def alter_topic(
        self,
        path: str,
        set_min_active_partitions: Optional[int] = None,
        set_max_active_partitions: Optional[int] = None,
        set_partition_count_limit: Optional[int] = None,
        add_consumers: Optional[List[Union[TopicConsumer, str]]] = None,
        alter_consumers: Optional[List[Union[TopicAlterConsumer, str]]] = None,
        drop_consumers: Optional[List[str]] = None,
        alter_attributes: Optional[Dict[str, str]] = None,
        set_metering_mode: Optional[TopicMeteringMode] = None,
        set_partition_write_speed_bytes_per_second: Optional[int] = None,
        set_partition_write_burst_bytes: Optional[int] = None,
        set_retention_period: Optional[datetime.timedelta] = None,
        set_retention_storage_mb: Optional[int] = None,
        set_supported_codecs: Optional[List[Union[TopicCodec, int]]] = None,
        alter_auto_partitioning_settings: Optional[TopicAlterAutoPartitioningSettings] = None,
    ):
        """
        alter topic command

        :param path: full path to topic
        :param set_min_active_partitions: Minimum partition count auto merge would stop working at.
        :param set_partition_count_limit: Limit for total partition count, including active (open for write)
            and read-only partitions.
        :param add_consumers: List of consumers for this topic to add
        :param alter_consumers: List of consumers for this topic to alter
        :param drop_consumers: List of consumer names for this topic to drop
        :param alter_attributes: User and server attributes of topic.
            Server attributes starts from "_" and will be validated by server.
        :param set_metering_mode: Metering mode for the topic in a serverless database
        :param set_partition_write_speed_bytes_per_second: Partition write speed in bytes per second
        :param set_partition_write_burst_bytes: Burst size for write in partition, in bytes
        :param set_retention_period: How long data in partition should be stored
        :param set_retention_storage_mb: How much data in partition should be stored
        :param set_supported_codecs: List of allowed codecs for writers. Writes with codec not from this list are forbidden.
            Empty list mean disable codec compatibility checks for the topic.
        """
        logger.debug("Alter topic request: path=%s", path)
        args = locals().copy()
        del args["self"]
        self._check_closed()

        req = _ydb_topic_public_types.AlterTopicRequestParams(**args)
        req = _ydb_topic.AlterTopicRequest.from_public(req)
        self._driver(
            req.to_proto(),
            _apis.TopicService.Stub,
            _apis.TopicService.AlterTopic,
            _wrap_operation,
        )

    def describe_topic(self, path: str, include_stats: bool = False) -> TopicDescription:
        logger.debug("Describe topic request: path=%s", path)
        args = locals().copy()
        del args["self"]
        self._check_closed()

        req = _ydb_topic_public_types.DescribeTopicRequestParams(**args)
        res = self._driver(
            req.to_proto(),
            _apis.TopicService.Stub,
            _apis.TopicService.DescribeTopic,
            _create_result_wrapper(_ydb_topic.DescribeTopicResult),
        )  # type: _ydb_topic.DescribeTopicResult
        return res.to_public()

    def drop_topic(self, path: str):
        self._check_closed()

        logger.debug("Drop topic request: path=%s", path)

        req = _ydb_topic_public_types.DropTopicRequestParams(path=path)
        self._driver(
            req.to_proto(),
            _apis.TopicService.Stub,
            _apis.TopicService.DropTopic,
            _wrap_operation,
        )

    def reader(
        self,
        topic: Union[str, TopicReaderSelector, List[Union[str, TopicReaderSelector]]],
        consumer: Optional[str],
        buffer_size_bytes: int = 50 * 1024 * 1024,
        # decoders: map[codec_code] func(encoded_bytes)->decoded_bytes
        # the func will be called from multiply threads in parallel
        decoders: Union[Mapping[int, Callable[[bytes], bytes]], None] = None,
        # custom decoder executor for call builtin and custom decoders. If None - use shared executor pool.
        # if max_worker in the executor is 1 - then decoders will be called from the thread without parallel
        decoder_executor: Optional[concurrent.futures.Executor] = None,  # default shared client executor pool
        auto_partitioning_support: Optional[bool] = True,  # Auto partitioning feature flag. Default - True.
        event_handler: Optional[TopicReaderEvents.EventHandler] = None,
    ) -> TopicReader:
        logger.debug("Create reader for topic=%s consumer=%s", topic, consumer)
        if not decoder_executor:
            decoder_executor = self._executor

        args = locals().copy()
        del args["self"]

        if consumer == "":
            raise issues.Error(
                "Consumer name could not be empty! To use reader without consumer specify consumer as None."
            )

        if consumer is None:
            if not isinstance(topic, TopicReaderSelector) or topic.partitions is None:
                raise issues.Error(
                    "To use reader without consumer it is required to specify partition_ids in topic selector."
                )

            if event_handler is None:
                raise issues.Error(
                    "To use reader without consumer it is required to specify event_handler with "
                    "on_partition_get_start_offset method."
                )

        settings = TopicReaderSettings(**args)

        return TopicReader(self._driver, settings, _parent=self)

    def writer(
        self,
        topic,
        *,
        producer_id: Optional[str] = None,  # default - random
        session_metadata: Mapping[str, str] = None,
        partition_id: Union[int, None] = None,
        auto_seqno: bool = True,
        auto_created_at: bool = True,
        codec: Optional[TopicCodec] = None,  # default mean auto-select
        # encoders: map[codec_code] func(encoded_bytes)->decoded_bytes
        # the func will be called from multiply threads in parallel.
        encoders: Optional[Mapping[_ydb_topic_public_types.PublicCodec, Callable[[bytes], bytes]]] = None,
        # custom encoder executor for call builtin and custom decoders. If None - use shared executor pool.
        # If max_worker in the executor is 1 - then encoders will be called from the thread without parallel.
        encoder_executor: Optional[concurrent.futures.Executor] = None,  # default shared client executor pool
    ) -> TopicWriter:
        logger.debug("Create writer for topic=%s producer_id=%s", topic, producer_id)
        args = locals().copy()
        del args["self"]
        self._check_closed()

        settings = TopicWriterSettings(**args)

        if not settings.encoder_executor:
            settings.encoder_executor = self._executor

        return TopicWriter(self._driver, settings, _parent=self)

    def tx_writer(
        self,
        tx,
        topic,
        *,
        producer_id: Optional[str] = None,  # default - random
        session_metadata: Mapping[str, str] = None,
        partition_id: Union[int, None] = None,
        auto_seqno: bool = True,
        auto_created_at: bool = True,
        codec: Optional[TopicCodec] = None,  # default mean auto-select
        # encoders: map[codec_code] func(encoded_bytes)->decoded_bytes
        # the func will be called from multiply threads in parallel.
        encoders: Optional[Mapping[_ydb_topic_public_types.PublicCodec, Callable[[bytes], bytes]]] = None,
        # custom encoder executor for call builtin and custom decoders. If None - use shared executor pool.
        # If max_worker in the executor is 1 - then encoders will be called from the thread without parallel.
        encoder_executor: Optional[concurrent.futures.Executor] = None,  # default shared client executor pool
    ) -> TopicWriter:
        logger.debug("Create tx writer for topic=%s tx=%s", topic, tx)
        args = locals().copy()
        del args["self"]
        del args["tx"]
        self._check_closed()

        settings = TopicWriterSettings(**args)

        if not settings.encoder_executor:
            settings.encoder_executor = self._executor

        return TopicTxWriter(tx, self._driver, settings, _parent=self)

    @ydb_retry(retry_cancelled=True, idempotent=True)
    def commit_offset(
        self, path: str, consumer: str, partition_id: int, offset: int, read_session_id: Optional[str] = None
    ) -> None:
        logger.debug(
            "Commit offset: path=%s partition_id=%s offset=%s consumer=%s",
            path,
            partition_id,
            offset,
            consumer,
        )
        req = _ydb_topic.CommitOffsetRequest(
            path=path,
            consumer=consumer,
            partition_id=partition_id,
            offset=offset,
            read_session_id=read_session_id,
        )

        self._driver(
            req.to_proto(),
            _apis.TopicService.Stub,
            _apis.TopicService.CommitOffset,
            _wrap_operation,
        )

    def close(self):
        if self._closed:
            return

        logger.debug("Close topic client")
        self._closed = True
        self._executor.shutdown(wait=False)
        logger.debug("Topic client was closed")

    def _check_closed(self):
        if not self._closed:
            return

        raise issues.Error("Topic client closed")


@dataclass
class TopicClientSettings:
    # ATTENTION
    # When set the encode_decode_threads_count - all custom encoders/decoders for topic reader/writer
    # MUST be thread-safe
    # because they will be called from parallel threads
    encode_decode_threads_count: int = 1


class TopicError(issues.Error):
    pass



import asyncio
from asyncio import wait_for
import pytest
from unittest import mock
import ydb

DEFAULT_TIMEOUT = 0.5
DEFAULT_RETRY_SETTINGS = ydb.RetrySettings(max_retries=1)


@pytest.mark.asyncio
class TestTopicTransactionalReader:
    async def test_commit(self, driver: ydb.aio.Driver, topic_with_messages, topic_consumer):
        async with ydb.aio.QuerySessionPool(driver) as pool:
            async with driver.topic_client.reader(topic_with_messages, topic_consumer) as reader:

                async def callee(tx: ydb.aio.QueryTxContext):
                    batch = await wait_for(reader.receive_batch_with_tx(tx, max_messages=1), DEFAULT_TIMEOUT)
                    assert len(batch.messages) == 1
                    assert batch.messages[0].data.decode() == "123"

                    batch = await wait_for(reader.receive_batch_with_tx(tx, max_messages=1), DEFAULT_TIMEOUT)
                    assert len(batch.messages) == 1
                    assert batch.messages[0].data.decode() == "456"

                await pool.retry_tx_async(callee, retry_settings=DEFAULT_RETRY_SETTINGS)
                assert len(reader._reconnector._tx_to_batches_map) == 0

            async with driver.topic_client.reader(topic_with_messages, topic_consumer) as reader:
                msg = await wait_for(reader.receive_message(), DEFAULT_TIMEOUT)
                assert msg.data.decode() == "789"

    async def test_rollback(self, driver: ydb.aio.Driver, topic_with_messages, topic_consumer):
        async with driver.topic_client.reader(topic_with_messages, topic_consumer) as reader:
            async with ydb.aio.QuerySessionPool(driver) as pool:

                async def callee(tx: ydb.aio.QueryTxContext):
                    batch = await wait_for(reader.receive_batch_with_tx(tx, max_messages=1), DEFAULT_TIMEOUT)
                    assert len(batch.messages) == 1
                    assert batch.messages[0].data.decode() == "123"

                    await tx.rollback()

                await pool.retry_tx_async(callee, retry_settings=DEFAULT_RETRY_SETTINGS)
                assert len(reader._reconnector._tx_to_batches_map) == 0

                msg = await wait_for(reader.receive_message(), DEFAULT_TIMEOUT)
                assert msg.data.decode() == "123"

    async def test_tx_failed_if_update_offsets_call_failed(
        self, driver: ydb.aio.Driver, topic_with_messages, topic_consumer
    ):
        async with driver.topic_client.reader(topic_with_messages, topic_consumer) as reader:
            async with ydb.aio.QuerySessionPool(driver) as pool:
                with mock.patch.object(
                    reader._reconnector,
                    "_do_commit_batches_with_tx_call",
                    side_effect=ydb.Error("Update offsets in tx failed"),
                ):

                    async def callee(tx: ydb.aio.QueryTxContext):
                        batch = await wait_for(reader.receive_batch_with_tx(tx, max_messages=1), DEFAULT_TIMEOUT)
                        assert len(batch.messages) == 1
                        assert batch.messages[0].data.decode() == "123"

                    with pytest.raises(ydb.Error, match="Transaction was failed"):
                        await pool.retry_tx_async(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

                    assert len(reader._reconnector._tx_to_batches_map) == 0

                    msg = await wait_for(reader.receive_message(), DEFAULT_TIMEOUT)
                    assert msg.data.decode() == "123"

    async def test_error_in_lambda(self, driver: ydb.aio.Driver, topic_with_messages, topic_consumer):
        async with driver.topic_client.reader(topic_with_messages, topic_consumer) as reader:
            async with ydb.aio.QuerySessionPool(driver) as pool:

                async def callee(tx: ydb.aio.QueryTxContext):
                    batch = await wait_for(reader.receive_batch_with_tx(tx, max_messages=1), DEFAULT_TIMEOUT)
                    assert len(batch.messages) == 1
                    assert batch.messages[0].data.decode() == "123"

                    raise RuntimeError("Something went wrong")

                with pytest.raises(RuntimeError):
                    await pool.retry_tx_async(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

                assert len(reader._reconnector._tx_to_batches_map) == 0

                msg = await wait_for(reader.receive_message(), DEFAULT_TIMEOUT)
                assert msg.data.decode() == "123"

    async def test_error_during_commit(self, driver: ydb.aio.Driver, topic_with_messages, topic_consumer):
        async with driver.topic_client.reader(topic_with_messages, topic_consumer) as reader:
            async with ydb.aio.QuerySessionPool(driver) as pool:

                async def callee(tx: ydb.aio.QueryTxContext):
                    with mock.patch.object(
                        tx,
                        "_commit_call",
                        side_effect=ydb.Unavailable("YDB Unavailable"),
                    ):
                        batch = await wait_for(reader.receive_batch_with_tx(tx, max_messages=1), DEFAULT_TIMEOUT)
                        assert len(batch.messages) == 1
                        assert batch.messages[0].data.decode() == "123"

                        await tx.commit()

                with pytest.raises(ydb.Unavailable):
                    await pool.retry_tx_async(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

                assert len(reader._reconnector._tx_to_batches_map) == 0

                msg = await wait_for(reader.receive_message(), DEFAULT_TIMEOUT)
                assert msg.data.decode() == "123"

    async def test_error_during_rollback(self, driver: ydb.aio.Driver, topic_with_messages, topic_consumer):
        async with driver.topic_client.reader(topic_with_messages, topic_consumer) as reader:
            async with ydb.aio.QuerySessionPool(driver) as pool:

                async def callee(tx: ydb.aio.QueryTxContext):
                    with mock.patch.object(
                        tx,
                        "_rollback_call",
                        side_effect=ydb.Unavailable("YDB Unavailable"),
                    ):
                        batch = await wait_for(reader.receive_batch_with_tx(tx, max_messages=1), DEFAULT_TIMEOUT)
                        assert len(batch.messages) == 1
                        assert batch.messages[0].data.decode() == "123"

                        await tx.rollback()

                with pytest.raises(ydb.Unavailable):
                    await pool.retry_tx_async(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

                assert len(reader._reconnector._tx_to_batches_map) == 0

                msg = await wait_for(reader.receive_message(), DEFAULT_TIMEOUT)
                assert msg.data.decode() == "123"


class TestTopicTransactionalReaderSync:
    def test_commit(self, driver_sync: ydb.Driver, topic_with_messages, topic_consumer):
        with ydb.QuerySessionPool(driver_sync) as pool:
            with driver_sync.topic_client.reader(topic_with_messages, topic_consumer) as reader:

                def callee(tx: ydb.QueryTxContext):
                    batch = reader.receive_batch_with_tx(tx, max_messages=1, timeout=DEFAULT_TIMEOUT)
                    assert len(batch.messages) == 1
                    assert batch.messages[0].data.decode() == "123"

                    batch = reader.receive_batch_with_tx(tx, max_messages=1, timeout=DEFAULT_TIMEOUT)
                    assert len(batch.messages) == 1
                    assert batch.messages[0].data.decode() == "456"

                pool.retry_tx_sync(callee, retry_settings=DEFAULT_RETRY_SETTINGS)
                assert len(reader._async_reader._reconnector._tx_to_batches_map) == 0

            with driver_sync.topic_client.reader(topic_with_messages, topic_consumer) as reader:
                msg = reader.receive_message(timeout=DEFAULT_TIMEOUT)
                assert msg.data.decode() == "789"

    def test_rollback(self, driver_sync: ydb.Driver, topic_with_messages, topic_consumer):
        with driver_sync.topic_client.reader(topic_with_messages, topic_consumer) as reader:
            with ydb.QuerySessionPool(driver_sync) as pool:

                def callee(tx: ydb.QueryTxContext):
                    batch = reader.receive_batch_with_tx(tx, max_messages=1, timeout=DEFAULT_TIMEOUT)
                    assert len(batch.messages) == 1
                    assert batch.messages[0].data.decode() == "123"

                    tx.rollback()

                pool.retry_tx_sync(callee, retry_settings=DEFAULT_RETRY_SETTINGS)
                assert len(reader._async_reader._reconnector._tx_to_batches_map) == 0

                msg = reader.receive_message(timeout=DEFAULT_TIMEOUT)
                assert msg.data.decode() == "123"

    def test_tx_failed_if_update_offsets_call_failed(
        self, driver_sync: ydb.Driver, topic_with_messages, topic_consumer
    ):
        with driver_sync.topic_client.reader(topic_with_messages, topic_consumer) as reader:
            with ydb.QuerySessionPool(driver_sync) as pool:
                with mock.patch.object(
                    reader._async_reader._reconnector,
                    "_do_commit_batches_with_tx_call",
                    side_effect=ydb.Error("Update offsets in tx failed"),
                ):

                    def callee(tx: ydb.QueryTxContext):
                        batch = reader.receive_batch_with_tx(tx, max_messages=1, timeout=DEFAULT_TIMEOUT)
                        assert len(batch.messages) == 1
                        assert batch.messages[0].data.decode() == "123"

                    with pytest.raises(ydb.Error, match="Transaction was failed"):
                        pool.retry_tx_sync(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

                    assert len(reader._async_reader._reconnector._tx_to_batches_map) == 0

                    msg = reader.receive_message(timeout=DEFAULT_TIMEOUT)
                    assert msg.data.decode() == "123"

    def test_error_in_lambda(self, driver_sync: ydb.Driver, topic_with_messages, topic_consumer):
        with driver_sync.topic_client.reader(topic_with_messages, topic_consumer) as reader:
            with ydb.QuerySessionPool(driver_sync) as pool:

                def callee(tx: ydb.QueryTxContext):
                    batch = reader.receive_batch_with_tx(tx, max_messages=1, timeout=DEFAULT_TIMEOUT)
                    assert len(batch.messages) == 1
                    assert batch.messages[0].data.decode() == "123"

                    raise RuntimeError("Something went wrong")

                with pytest.raises(RuntimeError):
                    pool.retry_tx_sync(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

                assert len(reader._async_reader._reconnector._tx_to_batches_map) == 0

                msg = reader.receive_message(timeout=DEFAULT_TIMEOUT)
                assert msg.data.decode() == "123"

    def test_error_during_commit(self, driver_sync: ydb.Driver, topic_with_messages, topic_consumer):
        with driver_sync.topic_client.reader(topic_with_messages, topic_consumer) as reader:
            with ydb.QuerySessionPool(driver_sync) as pool:

                def callee(tx: ydb.QueryTxContext):
                    with mock.patch.object(
                        tx,
                        "_commit_call",
                        side_effect=ydb.Unavailable("YDB Unavailable"),
                    ):
                        batch = reader.receive_batch_with_tx(tx, max_messages=1, timeout=DEFAULT_TIMEOUT)
                        assert len(batch.messages) == 1
                        assert batch.messages[0].data.decode() == "123"

                        tx.commit()

                with pytest.raises(ydb.Unavailable):
                    pool.retry_tx_sync(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

                assert len(reader._async_reader._reconnector._tx_to_batches_map) == 0

                msg = reader.receive_message(timeout=DEFAULT_TIMEOUT)
                assert msg.data.decode() == "123"

    def test_error_during_rollback(self, driver_sync: ydb.Driver, topic_with_messages, topic_consumer):
        with driver_sync.topic_client.reader(topic_with_messages, topic_consumer) as reader:
            with ydb.QuerySessionPool(driver_sync) as pool:

                def callee(tx: ydb.QueryTxContext):
                    with mock.patch.object(
                        tx,
                        "_rollback_call",
                        side_effect=ydb.Unavailable("YDB Unavailable"),
                    ):
                        batch = reader.receive_batch_with_tx(tx, max_messages=1, timeout=DEFAULT_TIMEOUT)
                        assert len(batch.messages) == 1
                        assert batch.messages[0].data.decode() == "123"

                        tx.rollback()

                with pytest.raises(ydb.Unavailable):
                    pool.retry_tx_sync(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

                assert len(reader._async_reader._reconnector._tx_to_batches_map) == 0

                msg = reader.receive_message(timeout=DEFAULT_TIMEOUT)
                assert msg.data.decode() == "123"


class TestTopicTransactionalWriter:
    async def test_commit(self, driver: ydb.aio.Driver, topic_path, topic_reader: ydb.TopicReaderAsyncIO):
        async with ydb.aio.QuerySessionPool(driver) as pool:

            async def callee(tx: ydb.aio.QueryTxContext):
                tx_writer = driver.topic_client.tx_writer(tx, topic_path)
                await tx_writer.write(ydb.TopicWriterMessage(data="123".encode()))

            await pool.retry_tx_async(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

        msg = await wait_for(topic_reader.receive_message(), 0.1)
        assert msg.data.decode() == "123"

    async def test_rollback(self, driver: ydb.aio.Driver, topic_path, topic_reader: ydb.TopicReaderAsyncIO):
        async with ydb.aio.QuerySessionPool(driver) as pool:

            async def callee(tx: ydb.aio.QueryTxContext):
                tx_writer = driver.topic_client.tx_writer(tx, topic_path)
                await tx_writer.write(ydb.TopicWriterMessage(data="123".encode()))

                await tx.rollback()

            await pool.retry_tx_async(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

        with pytest.raises(asyncio.TimeoutError):
            await wait_for(topic_reader.receive_message(), 0.1)

    async def test_no_msg_written_in_error_case(
        self, driver: ydb.aio.Driver, topic_path, topic_reader: ydb.TopicReaderAsyncIO
    ):
        async with ydb.aio.QuerySessionPool(driver) as pool:

            async def callee(tx: ydb.aio.QueryTxContext):
                tx_writer = driver.topic_client.tx_writer(tx, topic_path)
                await tx_writer.write(ydb.TopicWriterMessage(data="123".encode()))

                raise BaseException("error")

            with pytest.raises(BaseException):
                await pool.retry_tx_async(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

        with pytest.raises(asyncio.TimeoutError):
            await wait_for(topic_reader.receive_message(), 0.1)

    async def test_no_msg_written_in_tx_commit_error(
        self, driver: ydb.aio.Driver, topic_path, topic_reader: ydb.TopicReaderAsyncIO
    ):
        async with ydb.aio.QuerySessionPool(driver) as pool:

            async def callee(tx: ydb.aio.QueryTxContext):
                with mock.patch.object(
                    tx,
                    "_commit_call",
                    side_effect=ydb.Unavailable("YDB Unavailable"),
                ):
                    tx_writer = driver.topic_client.tx_writer(tx, topic_path)
                    await tx_writer.write(ydb.TopicWriterMessage(data="123".encode()))
                    await tx.commit()

            with pytest.raises(ydb.Unavailable):
                await pool.retry_tx_async(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

        with pytest.raises(asyncio.TimeoutError):
            await wait_for(topic_reader.receive_message(), 0.1)

    async def test_msg_written_exactly_once_with_retries(
        self, driver: ydb.aio.Driver, topic_path, topic_reader: ydb.TopicReaderAsyncIO
    ):
        error_raised = False
        async with ydb.aio.QuerySessionPool(driver) as pool:

            async def callee(tx: ydb.aio.QueryTxContext):
                nonlocal error_raised
                tx_writer = driver.topic_client.tx_writer(tx, topic_path)
                await tx_writer.write(ydb.TopicWriterMessage(data="123".encode()))

                if not error_raised:
                    error_raised = True
                    raise ydb.issues.Unavailable("some retriable error")

            await pool.retry_tx_async(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

            msg = await wait_for(topic_reader.receive_message(), 0.1)
            assert msg.data.decode() == "123"

            with pytest.raises(asyncio.TimeoutError):
                await wait_for(topic_reader.receive_message(), 0.1)

    async def test_writes_do_not_conflict_with_executes(self, driver: ydb.aio.Driver, topic_path):
        async with ydb.aio.QuerySessionPool(driver) as pool:

            async def callee(tx: ydb.aio.QueryTxContext):
                tx_writer = driver.topic_client.tx_writer(tx, topic_path)
                for _ in range(3):
                    async with await tx.execute("select 1"):
                        await tx_writer.write(ydb.TopicWriterMessage(data="123".encode()))

            await pool.retry_tx_async(callee, retry_settings=DEFAULT_RETRY_SETTINGS)


class TestTopicTransactionalWriterSync:
    def test_commit(self, driver_sync: ydb.Driver, topic_path, topic_reader_sync: ydb.TopicReader):
        with ydb.QuerySessionPool(driver_sync) as pool:

            def callee(tx: ydb.QueryTxContext):
                tx_writer = driver_sync.topic_client.tx_writer(tx, topic_path)
                tx_writer.write(ydb.TopicWriterMessage(data="123".encode()))

            pool.retry_tx_sync(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

        msg = topic_reader_sync.receive_message(timeout=DEFAULT_TIMEOUT)
        assert msg.data.decode() == "123"

    def test_rollback(self, driver_sync: ydb.Driver, topic_path, topic_reader_sync: ydb.TopicReader):
        with ydb.QuerySessionPool(driver_sync) as pool:

            def callee(tx: ydb.QueryTxContext):
                tx_writer = driver_sync.topic_client.tx_writer(tx, topic_path)
                tx_writer.write(ydb.TopicWriterMessage(data="123".encode()))

                tx.rollback()

            pool.retry_tx_sync(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

        with pytest.raises(TimeoutError):
            topic_reader_sync.receive_message(timeout=DEFAULT_TIMEOUT)

    def test_no_msg_written_in_error_case(
        self, driver_sync: ydb.Driver, topic_path, topic_reader_sync: ydb.TopicReader
    ):
        with ydb.QuerySessionPool(driver_sync) as pool:

            def callee(tx: ydb.QueryTxContext):
                tx_writer = driver_sync.topic_client.tx_writer(tx, topic_path)
                tx_writer.write(ydb.TopicWriterMessage(data="123".encode()))

                raise BaseException("error")

            with pytest.raises(BaseException):
                pool.retry_tx_sync(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

        with pytest.raises(TimeoutError):
            topic_reader_sync.receive_message(timeout=DEFAULT_TIMEOUT)

    def test_no_msg_written_in_tx_commit_error(
        self, driver_sync: ydb.Driver, topic_path, topic_reader_sync: ydb.TopicReader
    ):
        with ydb.QuerySessionPool(driver_sync) as pool:

            def callee(tx: ydb.QueryTxContext):
                with mock.patch.object(
                    tx,
                    "_commit_call",
                    side_effect=ydb.Unavailable("YDB Unavailable"),
                ):
                    tx_writer = driver_sync.topic_client.tx_writer(tx, topic_path)
                    tx_writer.write(ydb.TopicWriterMessage(data="123".encode()))
                    tx.commit()

            with pytest.raises(ydb.Unavailable):
                pool.retry_tx_sync(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

        with pytest.raises(TimeoutError):
            topic_reader_sync.receive_message(timeout=DEFAULT_TIMEOUT)

    def test_msg_written_exactly_once_with_retries(
        self, driver_sync: ydb.Driver, topic_path, topic_reader_sync: ydb.TopicReader
    ):
        error_raised = False
        with ydb.QuerySessionPool(driver_sync) as pool:

            def callee(tx: ydb.QueryTxContext):
                nonlocal error_raised
                tx_writer = driver_sync.topic_client.tx_writer(tx, topic_path)
                tx_writer.write(ydb.TopicWriterMessage(data="123".encode()))

                if not error_raised:
                    error_raised = True
                    raise ydb.issues.Unavailable("some retriable error")

            pool.retry_tx_sync(callee, retry_settings=DEFAULT_RETRY_SETTINGS)

            msg = topic_reader_sync.receive_message(timeout=DEFAULT_TIMEOUT)
            assert msg.data.decode() == "123"

            with pytest.raises(TimeoutError):
                topic_reader_sync.receive_message(timeout=DEFAULT_TIMEOUT)

    def test_writes_do_not_conflict_with_executes(self, driver_sync: ydb.Driver, topic_path):
        with ydb.QuerySessionPool(driver_sync) as pool:

            def callee(tx: ydb.QueryTxContext):
                tx_writer = driver_sync.topic_client.tx_writer(tx, topic_path)
                for _ in range(3):
                    with tx.execute("select 1"):
                        tx_writer.write(ydb.TopicWriterMessage(data="123".encode()))

            pool.retry_tx_sync(callee, retry_settings=DEFAULT_RETRY_SETTINGS)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: draft/ydb_object_storage_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v3.draft.protos import ydb_object_storage_pb2 as draft_dot_protos_dot_ydb__object__storage__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='draft/ydb_object_storage_v1.proto',
  package='Ydb.ObjectStorage.V1',
  syntax='proto3',
  serialized_options=b'\n&tech.ydb.proto.draft.object_storage.v1ZBgithub.com/ydb-platform/ydb-go-genproto/draft/Ydb_ObjectStorage_V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n!draft/ydb_object_storage_v1.proto\x12\x14Ydb.ObjectStorage.V1\x1a%draft/protos/ydb_object_storage.proto2e\n\x14ObjectStorageService\x12M\n\x04List\x12!.Ydb.ObjectStorage.ListingRequest\x1a\".Ydb.ObjectStorage.ListingResponseBl\n&tech.ydb.proto.draft.object_storage.v1ZBgithub.com/ydb-platform/ydb-go-genproto/draft/Ydb_ObjectStorage_V1b\x06proto3'
  ,
  dependencies=[draft_dot_protos_dot_ydb__object__storage__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None

_OBJECTSTORAGESERVICE = _descriptor.ServiceDescriptor(
  name='ObjectStorageService',
  full_name='Ydb.ObjectStorage.V1.ObjectStorageService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=98,
  serialized_end=199,
  methods=[
  _descriptor.MethodDescriptor(
    name='List',
    full_name='Ydb.ObjectStorage.V1.ObjectStorageService.List',
    index=0,
    containing_service=None,
    input_type=draft_dot_protos_dot_ydb__object__storage__pb2._LISTINGREQUEST,
    output_type=draft_dot_protos_dot_ydb__object__storage__pb2._LISTINGRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_OBJECTSTORAGESERVICE)

DESCRIPTOR.services_by_name['ObjectStorageService'] = _OBJECTSTORAGESERVICE

# @@protoc_insertion_point(module_scope)
